{-# LANGUAGE LambdaCase #-}


module Datact.Eval
  ( eval
  ) where


import Control.Applicative
import Control.Monad.Reader

import qualified Data.Map.Strict as Map


import Datact.Constants (Constants)
import Datact.Syntax


eval :: Constants -> Expr a -> Maybe Int
eval consts e = runReaderT (go e) consts
  where
    go = \case
        BinaryExpr (ArithBinOp binOp) l r _ -> case binOp of
            Add -> liftOp (+) l r
            Sub -> liftOp (-) l r
            Mul -> liftOp (*) l r
            Div -> liftOp div l r
        UnaryExpr (ArithUnOp Neg) e' _ -> negate <$> go e'
        IdentExpr name _ -> do
            env <- ask
            lift (Map.lookup name env)
        IntegerExpr i _ -> return (fromIntegral i)
        _ -> lift Nothing


    liftOp f l r = liftA2 f (go l) (go r)
