module Datact.SymbolTable
  ( SymbolTable
  , LocalName(..)
  , Range(..)

  , fromModel
  ) where


import Control.Lens

import           Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import           Data.Maybe      (mapMaybe)


import Datact.Constants (Constants)
import Datact.Eval
import Datact.Syntax


type SymbolTable = Map LocalName Range


data LocalName = LocalName
  { lnModuleName :: !Name
  , lnVarName    :: !Name
  } deriving(Eq, Ord, Show)


data Range = Range !Int !Int deriving (Eq, Ord, Show)


fromModel :: Constants -> Model a -> SymbolTable
fromModel consts (Model defs) =
    Map.unions . fmap (fromModule consts) .
    toListOf (traverse._ModuleDef) $ defs


fromModule :: Constants -> Module a -> SymbolTable
fromModule consts m =
    Map.fromList (mapMaybe (fromVarDecl consts (modName m)) (modVars m))


fromVarDecl :: Constants -> Name -> VarDecl a -> Maybe (LocalName, Range)
fromVarDecl consts moduleName (VarDecl name (IntVarType lower upper) _ _) =
    (,) (LocalName moduleName name) <$>
    (Range <$> eval consts lower <*> eval consts upper)
fromVarDecl _ _ _ = Nothing
