package common;

import java.util.BitSet;
import java.util.Iterator;
import java.util.PrimitiveIterator.OfInt;
import java.util.stream.IntStream;

public class BitSetTools
{
	public static BitSet asBitSet(final int... indices)
	{
		final BitSet result = new BitSet();
		for (int i : indices) {
			result.set(i);
		}
		return result;
	}

	public static BitSet asBitSet(final Iterable<Integer> indices)
	{
		Iterator<Integer> iter = indices.iterator();
		return (iter instanceof OfInt) ? asBitSet((OfInt) iter) : asBitSet(iter);
	}

	public static BitSet asBitSet(final IntStream indices)
	{
		return asBitSet(indices.iterator());
	}

	public static BitSet asBitSet(final Iterator<Integer> indices)
	{
		final BitSet result = new BitSet();
		while (indices.hasNext()) {
			result.set(indices.next());
		}
		return result;
	}

	public static BitSet asBitSet(final OfInt indices)
	{
		final BitSet result = new BitSet();
		while (indices.hasNext()) {
			result.set(indices.nextInt());
		}
		return result;
	}

	public static BitSet shiftDown(final BitSet indices, final int offset)
	{
		assert offset >= 0 : "positive offset expected";
		return indices.get(offset, indices.length());
	}

	public static BitSet shiftUp(final BitSet indices, final int offset)
	{
		assert offset >= 0 : "positive offset expected";
		return asBitSet(indices.stream().map(i -> i + offset));
	}

	public static boolean areNonEmpty(final Iterable<BitSet> sets)
	{
		for (BitSet set : sets) {
			if (set.isEmpty()) {
				return false;
			}
		}
		return true;
	}

	public static BitSet complement(final int toIndex, final BitSet indices)
	{
		final BitSet complement = (BitSet) indices.clone();
		complement.flip(0, toIndex);
		return complement;
	}

	public static BitSet minus(final BitSet set, final BitSet... sets)
	{
		final BitSet difference = (BitSet) set.clone();
		for (BitSet each : sets) {
			difference.andNot(each);
		}
		return difference;
	}

	public static BitSet union(final BitSet... sets)
	{
		BitSet union = new BitSet();
		for (BitSet set : sets) {
			union.or(set);
		}
		return union;
	}

	public static BitSet union(final Iterable<BitSet> sets)
	{
		return union(sets.iterator());
	}

	public static BitSet union(final Iterator<BitSet> sets)
	{
		BitSet union = new BitSet();
		while (sets.hasNext()) {
			union.or(sets.next());
		}
		return union;
	}

	public static BitSet intersect(BitSet set, BitSet... sets)
	{
		BitSet intersection = (BitSet) set.clone();
		for (BitSet each : sets) {
			intersection.and(each);
		}
		return intersection;
	}

	public static BitSet intersect(Iterable<BitSet> sets)
	{
		return intersect(sets.iterator());
	}

	public static BitSet intersect(Iterator<BitSet> sets)
	{
		if (! sets.hasNext()) {
			throw new IllegalArgumentException("Expected at least one set.");
		}
		BitSet intersection = sets.next();
		while (sets.hasNext()) {
			intersection.and(sets.next());
		}
		return intersection;
	}

	public static boolean isSubset(BitSet subset, BitSet superset)
	{
		return minus(subset, superset).isEmpty();
	}
}