//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package parser.ast;

import parser.visitor.*;
import prism.PrismLangException;

public class RewardStructItem extends ASTElement
{
	// Synchronising action:
	// * null = none (i.e. state reward)
	// * synch-list, matching transitions for transtion reward
	private SynchList synchs;

	private boolean openSynch = false;

	// Guard expression
	private Expression states;
	// Reward expression
	private Expression reward;
	
	// constructor

	@Deprecated
	public RewardStructItem(String a, Expression s, Expression r)
	{
		if (a != null) {
			synchs = new SynchList();
			if (!a.isEmpty()) {
				synchs.addSynch(a);
			}
		}
		openSynch = false;
		states = s;
		reward = r;
	}
	
	public RewardStructItem(Expression s, Expression r)
	{
		synchs = null;
		openSynch = false;
		states = s;
		reward = r;
	}

	public RewardStructItem(SynchList synchs, boolean openSynch, Expression s, Expression r)
	{
		this.synchs = synchs;
		this.openSynch = openSynch;
		states = s;
		reward = r;
	}

	// Set methods
	
	public void setSynchs(SynchList synchs)
	{
		this.synchs = synchs;
	}
	
	public void setOpenSynch(boolean value)
	{
		openSynch = value;
	}
	
	public void setStates(Expression e)
	{
		states = e;
	}
	
	public void setReward(Expression e)
	{
		reward = e;
	}
	
	// Get methods
	
	public SynchList getSynchs()
	{
		return synchs;
	}

	public boolean getOpenSynch()
	{
		return openSynch;
	}

	public Expression getStates()
	{
		return states;
	}
	
	public Expression getReward()
	{
		return reward;
	}
	
	/**
	 *	Returns whether this reward is a state (false) or transition (true) reward
	 */
	public boolean isTransitionReward()
	{
		return (synchs!=null);
	}

	public boolean matches(Object action)
	{
		if (action == null) {
			return synchs.cardinality() == 0;
		} else if (action instanceof simulator.MultiAction) {
			return ((simulator.MultiAction)action).matches(synchs, openSynch);
		} else if (action instanceof String) {
			String s = (String) action;
			if (s.isEmpty()) {
				return synchs.cardinality() == 0;
			}
			// s is a real action
			if (synchs.cardinality() == 0 && openSynch) {
				return true;
			}
			return (synchs.cardinality() == 1) && (synchs.getSynch(0).equals(s));
		}

		throw new IllegalArgumentException("Can not match transition reward against action class " + action.getClass());
	}

	// Methods required for ASTElement:
	
	/**
	 * Visitor method.
	 */
	public Object accept(ASTVisitor v) throws PrismLangException
	{
		return v.visit(this);
	}
	
	/**
	 * Convert to string.
	 */
	public String toString()
	{
		String s = "";
		
		if (synchs != null) {
			if (openSynch) {
				s += "]" + synchs + "[ ";
			} else {
				s += "[" + synchs + "] ";
			}
			
		}
		s += states + " : " + reward + ";";
		
		return s;
	}
	
	/**
	 * Perform a deep copy.
	 */
	public ASTElement deepCopy()
	{
		RewardStructItem ret = new RewardStructItem(synchs != null ? synchs.clone() : null, openSynch, states.deepCopy(), reward.deepCopy());
		ret.setPosition(this);
		return ret;
	}

}

//------------------------------------------------------------------------------
