package simulator;

import java.util.ArrayList;
import java.util.BitSet;

import common.IterableBitSet;
import parser.ast.SynchList;

public class MultiAction
{
	private BitSet actions;
	private ArrayList<String> actionNames;

	public MultiAction(BitSet actions, ArrayList<String> actionNames) {
		this.actions = actions;
		this.actionNames = actionNames;
	}

	public boolean matches(SynchList synchs, boolean open)
	{
		if (open) {
			for (int i : synchs.getSynchIndizes()) {
				if (!actions.get(i)) {
					return false;
				}
			}
			return true;
		} else {
			BitSet bs = synchs.asBitSet();
			return bs.equals(actions);
		}
	}

	public Iterable<Integer> iterator() {
		return IterableBitSet.getSetBits(actions);
	} 

	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result + ((actions == null) ? 0 : actions.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj)
	{
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		MultiAction other = (MultiAction) obj;
		if (actions == null) {
			if (other.actions != null)
				return false;
		} else if (!actions.equals(other.actions))
			return false;
		return true;
	}

	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		boolean first = true;
		for (int i : IterableBitSet.getSetBits(actions)) {
			if (!first)
				sb.append(",");
			first = false;
			sb.append(actionNames.get(i-1));
		}
		return sb.toString();
	}

}
