package nl.cwi.reo.templates;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import nl.cwi.reo.interpret.ports.Port;

// TODO: Auto-generated Javadoc
/**
 * The Class ReoTemplate.
 */
public final class ReoTemplate {

	/** The reofile. */
	private final String reofile;
	
	/** The Reo version. */
	private final String version;

	/** The packagename. */
	private final String packagename;

	/** The name. */
	private final String name;

	/** The ports. */
	private final Set<Port> ports;

	/** The components. */
	private final List<Component> components;

	/** Includes */
	private final List<String> includes;
	
	/** Extra settings */
	private final Map<String, Object> settings;

	/**
	 * Instantiates a new reo template.
	 *
	 * @param reofile
	 *            the reofile
	 * @param packagename
	 *            the packagename
	 * @param name
	 *            the name
	 * @param components
	 *            the components
	 * @param includes
	 *            a list of language specific includes
	 */
	public ReoTemplate(String reofile, String version, String packagename, String name, List<Component> components, List<String> includes, Map<String,Object> settings) {
		this.reofile = reofile;
		this.version = version;
		this.packagename = packagename;
		this.name = name;
		this.components = Collections.unmodifiableList(components);
		Set<Port> P = new LinkedHashSet<Port>();
		for (Component c : components)
			P.addAll(c.getPorts());
		this.ports = Collections.unmodifiableSet(P);
		this.includes = Collections.unmodifiableList(includes);
		this.settings = settings;
	}

	/**
	 * Gets the file.
	 *
	 * @return the file
	 */
	public String getFile() {
		return reofile;
	}
	
	/**
	 * Gets the Reo version.
	 *
	 * @return the Reo version
	 */
	public String getVersion() {
		return version;
	}
	
	public Map<String, Object> getSettings() {
		return settings;
	}
	
	public List<String> getIncludes() {
		return includes;
	}
	
	/**
	 * Gets the package.
	 *
	 * @return the package
	 */
	public String getPackage() {
		return packagename;
	}

	/**
	 * Gets the name.
	 *
	 * @return the name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gets the ports.
	 *
	 * @return the ports
	 */
	public Set<Port> getPorts() {
		return ports;
	}

	/**
	 * Gets the components.
	 *
	 * @return the components
	 */
	public List<Component> getComponents() {
		return components;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		String s = name + ports + "\n";
		for (Component c : components)
			s += c.toString() + "\n";
		return s;
	}
}
