package nl.cwi.reo.semantics.automata;

import java.util.Objects;

import org.checkerframework.checker.nullness.qual.Nullable;

// TODO: Auto-generated Javadoc
/**
 * Container to ease passing around a tuple of two objects. This object provides
 * a sensible implementation of equals(), returning true if equals() is true on
 * each of the contained objects.
 *
 * @author Android SDK
 * @param <F>
 *            the generic type
 * @param <S>
 *            the generic type
 */
public class Pair<F, S> {

	/** The first. */
	public final F first;

	/** The second. */
	public final S second;

	/**
	 * Constructor for a Pair.
	 *
	 * @param first
	 *            the first object in the Pair
	 * @param second
	 *            the second object in the pair
	 */
	public Pair(F first, S second) {
		this.first = first;
		this.second = second;
	}

	/**
	 * Checks the two objects for equality by delegating to their respective
	 * {@link Object#equals(Object)} methods.
	 *
	 * @param o
	 *            the {@link Pair} to which this one is to be checked for
	 *            equality
	 * @return true if the underlying objects of the Pair are both considered
	 *         equal
	 */
	@Override
	public boolean equals(@Nullable Object o) {
		if (!(o instanceof Pair)) {
			return false;
		}
		Pair<?, ?> p = (Pair<?, ?>) o;
		return Objects.equals(p.first, first) && Objects.equals(p.second, second);
	}

	/**
	 * Compute a hash code using the hash codes of the underlying objects.
	 *
	 * @return a hashcode of the Pair
	 */
	@Override
	public int hashCode() {
		return Objects.hash(first, second);
	}

	/**
	 * Convenience method for creating an appropriately typed pair.
	 *
	 * @param <A>
	 *            the generic type
	 * @param <B>
	 *            the generic type
	 * @param a
	 *            the first object in the Pair
	 * @param b
	 *            the second object in the pair
	 * @return a Pair that is templatized with the types of a and b
	 */
	public static <A, B> Pair<A, B> create(A a, B b) {
		return new Pair<A, B>(a, b);
	}
}