package nl.cwi.reo.semantics.prba;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import nl.cwi.reo.interpret.ReoParser.Rba_distributionContext;
import nl.cwi.reo.interpret.ReoParser.RewContext;
import nl.cwi.reo.interpret.ReoParser.Rew_portContext;
import nl.cwi.reo.interpret.ReoParser.Rew_ruleContext;
import nl.cwi.reo.interpret.ports.Port;
import nl.cwi.reo.semantics.prba.RewardMonitor.Rule;
import nl.cwi.reo.semantics.predicates.Term;
import nl.cwi.reo.semantics.rulebasedautomata.ListenerRBA;
import nl.cwi.reo.util.Monitor;

/**
 * Listener for probabilistic rule-based automata.
 */
public class ListenerPRBA extends ListenerRBA {

	/**
	 * Instantiates a new listener PRBA.
	 *
	 * @param m
	 *            the monitor
	 */
	public ListenerPRBA(Monitor m) {
		super(m);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void exitRba_distribution(Rba_distributionContext ctx) {
		Map<Term, Term> distr = new LinkedHashMap<>();
		for (int i = 0; i < ctx.rba_term().size(); i += 2) {
			Term t1 = terms.get(ctx.rba_term(i));
			Term t2 = terms.get(ctx.rba_term(i+1));
			distr.put(t2, t1);
		}
		terms.put(ctx, new Distribution(distr));
	}
	
	@Override
	public void exitRew(RewContext ctx)
	{
		String globalName = null;
		if (ctx.STRING() != null) {
			globalName = ctx.STRING().getText();
			globalName = globalName.substring(1, globalName.length()-1);
		}
		List<Rule> rules = new ArrayList<>();
		for (Rew_ruleContext rule_ctx : ctx.rew_rule()) {
			List<Port> ports = new ArrayList<>();
			for (Rew_portContext port_ctx : rule_ctx.rew_ports().rew_port()) {
				Port p = new Port(port_ctx.getText());
				ports.add(p);
			}
			int value = Integer.parseInt(rule_ctx.rew_expr().getText());
			rules.add(new RewardMonitor.Rule(ports, value));
		}

		atoms.put(ctx, new RewardMonitor(globalName, rules));
	}

}
